"""Test the ping command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("ping_failure_response.json").encode()]
)
async def test_ping_failure(mock_datagram_client: MagicMock) -> None:
    """Test a failed ping of the device.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.system.ping()

        assert str(err.value) == (
            "SYSTEM_PING command failed (response: {'command': 0, 'status': 'error'})"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("ping_success_response.json").encode()]
)
async def test_ping_success(mock_datagram_client: MagicMock) -> None:
    """Test the ping command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            ping_response = await client.system.ping()

        assert ping_response["command"] == 0
        assert ping_response["status"] == "ok"
        assert ping_response["data"] == {"uid": "ABCDEF123456"}


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("ping_success_silent_response.json").encode()]
)
async def test_ping_silent_success(mock_datagram_client: MagicMock) -> None:
    """Test the ping command succeeding while silent.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            ping_response = await client.system.ping()

        assert ping_response["command"] == 0
        assert ping_response["status"] == "ok"
        assert ping_response["silent"] is True
        assert ping_response["data"] == {"uid": "ABCDEF123456"}



"""Test the reboot command."""
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("reboot_failure_response.json").encode()]
)
async def test_reboot_failure(mock_datagram_client: MagicMock) -> None:
    """Test the reboot command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.system.reboot()

        assert str(err.value) == (
            "SYSTEM_REBOOT command failed (response: {'command': 2, 'status': 'error'})"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("reboot_success_response.json").encode()]
)
async def test_reboot_success(mock_datagram_client: MagicMock) -> None:
    """Test the reboot command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            # Patch asyncio.sleep so that this test doesn't take 3-ish seconds:
            with patch("asyncio.sleep", AsyncMock()):
                reboot_response = await client.system.reboot()

        assert reboot_response["command"] == 2
        assert reboot_response["status"] == "ok"



"""Define docs configuration."""
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
from datetime import datetime

# -- Project information -----------------------------------------------------

project = "aioguardian"
copyright = f"{datetime.today().year}, Aaron Bach"
author = "Aaron Bach"

# The full version, including alpha/beta/rc tags
release = "2023.12.0"


# -- General configuration ---------------------------------------------------

master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.napoleon",
    "sphinx_rtd_theme",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {"display_version": True, "style_external_links": True}

# -- Options for autodoc -------------------------------------------------

autodoc_default_options = {"member-order": "bysource"}
autosectionlabel_prefix_document = True



"""Test the publish_state command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response",
    [load_fixture("publish_state_failure_response.json").encode()],
)
async def test_publish_state_failure(mock_datagram_client: MagicMock) -> None:
    """Test the publish_state command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client, pytest.raises(CommandError) as err:
        async with Client("192.168.1.100") as client:
            _ = await client.iot.publish_state()
        client.disconnect()

    assert str(err.value) == (
        "IOT_PUBLISH_STATE command failed "
        "(response: {'command': 65, 'status': 'error'})"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response",
    [load_fixture("publish_state_success_response.json").encode()],
)
async def test_publish_state_success(mock_datagram_client: MagicMock) -> None:
    """Test the publish_state command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            publish_state_response = await client.iot.publish_state()

        assert publish_state_response["command"] == 65
        assert publish_state_response["status"] == "ok"



"""Define generic fixtures for tests."""
# pylint: disable=redefined-outer-name
from collections.abc import Generator
from unittest.mock import AsyncMock, MagicMock, patch

import pytest


@pytest.fixture()
def command_response() -> MagicMock:
    """Define a fixture for the JSON response payload of a command.

    Returns:
        A mocked command response.
    """
    return MagicMock()


@pytest.fixture()
def mock_datagram_client(recv_response: AsyncMock) -> Generator:
    """Define a mocked datagram client.

    Args:
        recv_response: A mocked instance of a socket recv operation.
    """
    mock_datagram_client = MagicMock()
    mock_datagram_client.connect = AsyncMock()
    mock_datagram_client.recv = recv_response
    mock_datagram_client.send = AsyncMock()
    mock_datagram_client.close = MagicMock()

    with patch("asyncio_dgram.connect", return_value=mock_datagram_client):
        yield mock_datagram_client


@pytest.fixture()
def recv_response(command_response: MagicMock, remote_addr_response: str) -> AsyncMock:
    """Define a response from the socket.

    Returns:
        A mocked instance of a socket recv operation.
    """
    return AsyncMock(return_value=(command_response, remote_addr_response))


@pytest.fixture()
def remote_addr_response() -> str:
    """Define an IP address for the Guardian device.

    Returns:
        A response for getting the valve's remote IP address.
    """
    return "192.168.1.100"



"""Test the disable_ap command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("wifi_disable_ap_failure_response.json").encode()]
)
async def test_disable_ap_failure(mock_datagram_client: MagicMock) -> None:
    """Test the wifi_disable_ap command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.wifi.disable_ap()

        assert str(err.value) == (
            "WIFI_DISABLE_AP command failed "
            "(response: {'command': 36, 'status': 'error'})"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("wifi_disable_ap_success_response.json").encode()]
)
async def test_disable_ap_success(mock_datagram_client: MagicMock) -> None:
    """Test the wifi_disable_ap command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            wifi_disable_ap_response = await client.wifi.disable_ap()

        assert wifi_disable_ap_response["command"] == 36
        assert wifi_disable_ap_response["status"] == "ok"



"""Define valve commands."""
from collections.abc import Awaitable, Callable
from typing import Any

from aioguardian.const import LOGGER
from aioguardian.helpers.command import Command

VALVE_STATE_MAPPING = {
    0: "default",
    1: "start_opening",
    2: "opening",
    3: "finish_opening",
    4: "opened",
    5: "start_closing",
    6: "closing",
    7: "finish_closing",
    8: "closed",
    9: "start_halt",
    10: "stalled",
    11: "free_spin",
    12: "halted",
}


class ValveCommands:
    """Define an object to manage valve commands.

    Note that this class shouldn't be instantiated directly; an instance of it will
    automatically be added to the :meth:`Client <aioguardian.Client>` (as
    ``client.valve``).

    Args:
        execute_command: The execute_command method from the Client object.
    """

    def __init__(
        self, execute_command: Callable[..., Awaitable[dict[str, Any]]]
    ) -> None:
        """Initialize.

        Args:
            execute_command: The execute_command method from the Client object.
        """
        self._execute_command = execute_command

    async def close(self) -> dict[str, Any]:
        """Close the valve.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.VALVE_CLOSE)

    async def halt(self) -> dict[str, Any]:
        """Halt the valve.

        Note that calling this method will cause the device to no longer respond to leak
        events; therefore, it is not recommended to leave the device in a "halted" state
        indefinitely.

        Returns:
            An API response payload.
        """
        LOGGER.warning(
            "The device will not respond to leak events while in a halted state. It is "
            "recommended that you call valve_close() or valve_open() as soon as "
            "possible."
        )
        return await self._execute_command(Command.VALVE_HALT)

    async def open(self) -> dict[str, Any]:
        """Open the valve.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.VALVE_OPEN)

    async def reset(self, *, silent: bool = True) -> dict[str, Any]:
        """Reset the valve.

        This fully resets system motor diagnostics (including open/close count and
        lifetime average current draw) and cannot be undone.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.VALVE_RESET, silent=silent)

    async def status(self, *, silent: bool = True) -> dict[str, Any]:
        """Retrieve status of the valve.

        In the payload that is returned, the ``state`` attribute of the valve can be any
        one of the following:

            * ``closed``
            * ``closing``
            * ``default``
            * ``finish_closing``
            * ``finish_opening``
            * ``free_spin``
            * ``halted``
            * ``opened``
            * ``opening``
            * ``stalled``
            * ``start_closing``
            * ``start_halt``
            * ``start_opening``

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        data = await self._execute_command(Command.VALVE_STATUS, silent=silent)
        data["data"]["state"] = VALVE_STATE_MAPPING[data["data"]["state"]]
        return data



"""Define system commands."""
from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable
from typing import Any

import voluptuous as vol

import aioguardian.helpers.config_validation as cv
from aioguardian.errors import CommandError
from aioguardian.helpers.command import Command

PARAM_FILENAME = "filename"
PARAM_PORT = "port"
PARAM_URL = "url"

UPGRADE_FIRMWARE_PARAM_SCHEMA = vol.Schema(
    {
        vol.Optional(PARAM_URL): vol.All(cv.url, vol.Length(max=256)),
        vol.Optional(PARAM_PORT): int,
        vol.Optional(PARAM_FILENAME): vol.All(str, vol.Length(max=48)),
    }
)


class SystemCommands:
    """Define an object to manage system commands.

    Note that this class shouldn't be instantiated directly; an instance of it will
    automatically be added to the :meth:`Client <aioguardian.Client>` (as
    ``client.system``).

    Args:
        execute_command: The execute_command method from the Client object.
    """

    def __init__(
        self, execute_command: Callable[..., Awaitable[dict[str, Any]]]
    ) -> None:
        """Initialize.

        Args:
            execute_command: The execute_command method from the Client object.
        """
        self._execute_command = execute_command

    async def diagnostics(self, *, silent: bool = True) -> dict[str, Any]:
        """Retrieve diagnostics info.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.SYSTEM_DIAGNOSTICS, silent=silent)

    async def factory_reset(self, *, silent: bool = True) -> dict[str, Any]:
        """Perform a factory reset on the device.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.SYSTEM_FACTORY_RESET, silent=silent)

    async def onboard_sensor_status(self, *, silent: bool = True) -> dict[str, Any]:
        """Retrieve status of the valve controller's onboard sensors.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(
            Command.SYSTEM_ONBOARD_SENSOR_STATUS, silent=silent
        )

    async def ping(self, *, silent: bool = True) -> dict[str, Any]:
        """Ping the device.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.SYSTEM_PING, silent=silent)

    async def reboot(self, *, silent: bool = True) -> dict[str, Any]:
        """Reboot the device.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        data = await self._execute_command(Command.SYSTEM_REBOOT, silent=silent)

        # The Guardian API docs indicate that the reboot will occur "3 seconds after
        # command is received" – in order to guard against errors from subsequent
        # commands while the reboot is occurring, we sleep for 3 seconds before
        # returning the response:
        await asyncio.sleep(3)

        return data

    async def upgrade_firmware(
        self,
        *,
        url: str | None = None,
        port: int | None = None,
        filename: str | None = None,
        silent: bool = True,
    ) -> dict[str, Any]:
        """Upgrade the firmware on the device.

        Args:
            url: A Guardian-provided URL that hosts firmware files.
            port: A port at the Guardian-provided URL that can be accessed.
            filename: A firmware filename.
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.

        Raises:
            CommandError: Raised when invalid parameters are provided.
        """
        params: dict[str, int | str] = {}
        if url:
            params["url"] = url
        if port:
            params["port"] = port
        if filename:
            params["filename"] = filename

        try:
            UPGRADE_FIRMWARE_PARAM_SCHEMA(params)
        except vol.Invalid as err:
            raise CommandError(f"Invalid parameters provided: {err}") from None

        return await self._execute_command(
            Command.SYSTEM_UPGRADE_FIRMWARE, params=params, silent=silent
        )



"""Define sensor commands."""
from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Any

import voluptuous as vol

import aioguardian.helpers.config_validation as cv
from aioguardian.errors import CommandError
from aioguardian.helpers.command import Command

PARAM_UID = "uid"

PAIRED_SENSOR_UID_SCHEMA: vol.Schema = vol.Schema(
    {vol.Required(PARAM_UID): vol.All(cv.alphanumeric, vol.Length(max=12))}
)


class SensorCommands:
    """Define an object to manage sensor commands.

    Note that this class shouldn't be instantiated directly; an instance of it will
    automatically be added to the :meth:`Client <aioguardian.Client>` (as
    ``client.sensor``).

    Args:
        execute_command: The execute_command method from the Client object.
    """

    def __init__(
        self, execute_command: Callable[..., Awaitable[dict[str, Any]]]
    ) -> None:
        """Initialize.

        Args:
            execute_command: The execute_command method from the Client object.
        """
        self._execute_command = execute_command

    async def pair_dump(self, *, silent: bool = True) -> dict[str, Any]:
        """Dump information on all paired sensors.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.SENSOR_PAIR_DUMP, silent=silent)

    async def pair_sensor(self, uid: str, *, silent: bool = True) -> dict[str, Any]:
        """Pair a new sensor to the device.

        Args:
            uid: A UID of a Guardian paired sensor.
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.

        Raises:
            CommandError: Raised when invalid parameters are provided.
        """
        params = {PARAM_UID: uid}

        try:
            PAIRED_SENSOR_UID_SCHEMA(params)
        except vol.Invalid as err:
            raise CommandError(f"Invalid parameters provided: {err}") from None

        return await self._execute_command(
            Command.SENSOR_PAIR_SENSOR, params=params, silent=silent
        )

    async def paired_sensor_status(
        self, uid: str, *, silent: bool = True
    ) -> dict[str, Any]:
        """Get the status (leak, temperature, etc.) of a paired sensor.

        Args:
            uid: A UID of a Guardian paired sensor.
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.

        Raises:
            CommandError: Raised when invalid parameters are provided.
        """
        params = {PARAM_UID: uid}

        try:
            PAIRED_SENSOR_UID_SCHEMA(params)
        except vol.Invalid as err:
            raise CommandError(f"Invalid parameters provided: {err}") from None

        return await self._execute_command(
            Command.SENSOR_PAIRED_SENSOR_STATUS, params=params, silent=silent
        )

    async def unpair_sensor(self, uid: str, *, silent: bool = True) -> dict[str, Any]:
        """Unpair a sensor from the device.

        Args:
            uid: A UID of a Guardian paired sensor.
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.

        Raises:
            CommandError: Raised when invalid parameters are provided.
        """
        params = {PARAM_UID: uid}

        try:
            PAIRED_SENSOR_UID_SCHEMA(params)
        except vol.Invalid as err:
            raise CommandError(f"Invalid parameters provided: {err}") from None

        return await self._execute_command(
            Command.SENSOR_UNPAIR_SENSOR, params=params, silent=silent
        )



"""Test the status command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("valve_status_failure_response.json").encode()]
)
async def test_status_failure(mock_datagram_client: MagicMock) -> None:
    """Test the valve_status command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.valve.status()

        assert str(err.value) == (
            "VALVE_STATUS command failed (response: {'command': 16, 'status': 'error'})"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("valve_status_success_response.json").encode()]
)
async def test_status_success(mock_datagram_client: MagicMock) -> None:
    """Test the valve_status command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            valve_status_response = await client.valve.status()

        assert valve_status_response["command"] == 16
        assert valve_status_response["status"] == "ok"
        assert valve_status_response["data"] == {
            "enabled": False,
            "direction": True,
            "state": "default",
            "travel_count": 0,
            "instantaneous_current": 0,
            "instantaneous_current_ddt": 0,
            "average_current": 34,
        }



"""Test the configure command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError, GuardianError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("wifi_configure_failure_response.json").encode()]
)
async def test_configure_failure(mock_datagram_client: MagicMock) -> None:
    """Test the wifi_configure command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.wifi.configure("My_Network", "password123")

        assert str(err.value) == (
            "WIFI_CONFIGURE command failed "
            "(response: {'command': 34, 'status': 'error'})"
        )


@pytest.mark.asyncio
async def test_configure_invalid_password(mock_datagram_client: MagicMock) -> None:
    """Test that an invalid password throws an exception.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(GuardianError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.wifi.configure(
                    "My_Network",
                    "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
                )

        assert str(err.value) == (
            "Invalid parameters provided: length of value must be at most 64 for "
            "dictionary value @ data['password']"
        )


@pytest.mark.asyncio
async def test_configure_invalid_ssid(mock_datagram_client: MagicMock) -> None:
    """Test that an invalid SSID throws an exception.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(GuardianError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.wifi.configure(
                    "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx", "password123"
                )

        assert str(err.value) == (
            "Invalid parameters provided: length of value must be at most 36 for "
            "dictionary value @ data['ssid']"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("wifi_configure_success_response.json").encode()]
)
async def test_configure_success(mock_datagram_client: MagicMock) -> None:
    """Test the wifi_configure command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            wifi_configure_response = await client.wifi.configure(
                "My_Network", "password123"
            )

        assert wifi_configure_response["command"] == 34
        assert wifi_configure_response["status"] == "ok"



"""Test command helpers."""
import pytest

from aioguardian.errors import CommandError
from aioguardian.helpers.command import (
    Command,
    get_command_from_code,
    get_command_from_name,
)


@pytest.mark.asyncio
async def test_get_command_from_code() -> None:
    """Test the get_command_from_code helper."""
    real_command = get_command_from_code(0)
    assert real_command == Command.SYSTEM_PING

    with pytest.raises(CommandError) as err:
        _ = get_command_from_code(99999)
    assert str(err.value) == "Unknown command code: 99999"


@pytest.mark.asyncio
async def test_get_command_from_name() -> None:
    """Test the get_command_from_name helper."""
    real_command = get_command_from_name("SYSTEM_PING")
    assert real_command == Command.SYSTEM_PING

    with pytest.raises(CommandError) as err:
        _ = get_command_from_name("not real")
    assert str(err.value) == "Unknown command name: not real"



"""Define IOT commands."""
from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Any

from aioguardian.helpers.command import Command


class IOTCommands:  # pylint: disable=too-few-public-methods
    """Define an object to manage IOT commands.

    Note that this class shouldn't be instantiated directly; an instance of it will
    automatically be added to the :meth:`Client <aioguardian.Client>` (as
    ``client.iot``).

    Args:
        execute_command: The execute_command method from the Client object.
    """

    def __init__(
        self, execute_command: Callable[..., Awaitable[dict[str, Any]]]
    ) -> None:
        """Initialize.

        Args:
            execute_command: The execute_command method from the Client object.
        """
        self._execute_command = execute_command

    async def publish_state(self, *, silent: bool = True) -> dict[str, Any]:
        """Publish the device's complete state to the Guardian cloud.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.IOT_PUBLISH_STATE, silent=silent)



"""Test the status command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("wifi_status_failure_response.json").encode()]
)
async def test_status_failure(mock_datagram_client: MagicMock) -> None:
    """Test the wifi_status command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.wifi.status()

        assert str(err.value) == (
            "WIFI_STATUS command failed (response: {'command': 32, 'status': 'error'})"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("wifi_status_success_response.json").encode()]
)
async def test_status_success(mock_datagram_client: MagicMock) -> None:
    """Test the wifi_status command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            wifi_status_response = await client.wifi.status()

        assert wifi_status_response["command"] == 32
        assert wifi_status_response["status"] == "ok"
        assert wifi_status_response["data"] == {
            "station_connected": True,
            "ip_assigned": True,
            "mqtt_connected": True,
            "rssi": -63,
            "channel": 1,
            "lan_ipv4": "192.168.1.100",
            "lan_ipv6": "AC10:BD0:FFFF:FFFF:AC10:BD0:FFFF:FFFF",
            "ap_enabled": True,
            "ap_clients": 0,
            "bssid": "ABCDEF123456",
            "ssid": "My_Network",
        }



"""Define command helpers."""
from enum import Enum

from aioguardian.errors import CommandError


class Command(Enum):
    """Define a Guardian UDP command mapping."""

    IOT_PUBLISH_STATE = 65
    SENSOR_PAIR_DUMP = 48
    SENSOR_PAIR_SENSOR = 49
    SENSOR_PAIRED_SENSOR_STATUS = 51
    SENSOR_UNPAIR_SENSOR = 50
    SYSTEM_DIAGNOSTICS = 1
    SYSTEM_FACTORY_RESET = 255
    SYSTEM_ONBOARD_SENSOR_STATUS = 80
    SYSTEM_PING = 0
    SYSTEM_REBOOT = 2
    SYSTEM_UPGRADE_FIRMWARE = 4
    VALVE_CLOSE = 18
    VALVE_HALT = 19
    VALVE_OPEN = 17
    VALVE_RESET = 20
    VALVE_STATUS = 16
    WIFI_CONFIGURE = 34
    WIFI_DISABLE_AP = 36
    WIFI_ENABLE_AP = 35
    WIFI_LIST = 38
    WIFI_RESET = 33
    WIFI_SCAN = 37
    WIFI_STATUS = 32


def get_command_from_name(command_name: str) -> Command:
    """Return the command for a particular name.

    Args:
        command_name: The command string to parse into a Command.

    Returns:
        A Command object.

    Raises:
        CommandError: Raised when an unknown command is encountered.
    """
    try:
        command = Command[command_name]
    except KeyError:
        raise CommandError(f"Unknown command name: {command_name}") from None
    return command


def get_command_from_code(command_code: int) -> Command:
    """Return the command for a particular code.

    Args:
        command_code: The raw command code to parse into a Command.

    Returns:
        A Command object.

    Raises:
        CommandError: Raised when an unknown command is encountered.
    """
    try:
        command = Command(command_code)
    except ValueError:
        raise CommandError(f"Unknown command code: {command_code}") from None
    return command



"""Test the close command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("valve_close_failure_response.json").encode()]
)
async def test_close_failure(mock_datagram_client: MagicMock) -> None:
    """Test the valve_close command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.valve.close()

        assert str(err.value) == "VALVE_CLOSE command failed: valve_already_closed"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("valve_close_success_response.json").encode()]
)
async def test_close_success(mock_datagram_client: MagicMock) -> None:
    """Test the valve_close command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            valve_close_response = await client.valve.close()

        assert valve_close_response["command"] == 18



"""Test the upgrade_firmware command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import GuardianError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response",
    [load_fixture("upgrade_firmware_success_response.json").encode()],
)
async def test_upgrade_firmware_custom_parameters(
    mock_datagram_client: MagicMock,
) -> None:
    """Test that valid custom parameters work.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            upgrade_firmware_response = await client.system.upgrade_firmware(
                url="https://firmware.com", port=8080, filename="123.bin"
            )

        assert upgrade_firmware_response["command"] == 4
        assert upgrade_firmware_response["status"] == "ok"


@pytest.mark.asyncio
async def test_upgrade_firmware_invalid_filename(
    mock_datagram_client: MagicMock,
) -> None:
    """Test that an invalid firmware filename throws an exception.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(GuardianError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.system.upgrade_firmware(
                    filename="xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
                )

        assert str(err.value) == (
            "Invalid parameters provided: length of value must be at most 48 for "
            "dictionary value @ data['filename']"
        )


@pytest.mark.asyncio
async def test_upgrade_firmware_invalid_port(mock_datagram_client: MagicMock) -> None:
    """Test that an invalid firmware port throws an exception.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(GuardianError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.system.upgrade_firmware(
                    port="WHOOPS"  # type: ignore[arg-type]
                )

        assert str(err.value) == (
            "Invalid parameters provided: expected int for dictionary value @ "
            "data['port']"
        )


@pytest.mark.asyncio
async def test_upgrade_firmware_invalid_url(mock_datagram_client: MagicMock) -> None:
    """Test that an invalid firmware URL throws an exception.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(GuardianError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.system.upgrade_firmware(url="not_real_url")

        assert str(err.value) == (
            "Invalid parameters provided: Invalid URL for dictionary value @ "
            "data['url']"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response",
    [load_fixture("upgrade_firmware_success_response.json").encode()],
)
async def test_upgrade_firmware_success(mock_datagram_client: MagicMock) -> None:
    """Test the upgrade_firmware command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            upgrade_firmware_response = await client.system.upgrade_firmware()

        assert upgrade_firmware_response["command"] == 4
        assert upgrade_firmware_response["status"] == "ok"



"""Test the paired_sensor_status command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response",
    [load_fixture("paired_sensor_status_failure_3_response.json").encode()],
)
async def test_paired_sensor_status_failure_not_paired(
    mock_datagram_client: MagicMock,
) -> None:
    """Test the paired_sensor_status command failing because it isn't paired.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.sensor.paired_sensor_status("ABCDE1234567")

        assert (
            str(err.value)
            == "SENSOR_PAIRED_SENSOR_STATUS command failed: sensor_not_paired"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response",
    [load_fixture("paired_sensor_status_failure_5_response.json").encode()],
)
async def test_paired_sensor_status_failure_error_loading(
    mock_datagram_client: MagicMock,
) -> None:
    """Test the paired_sensor_status command failing because of a loading error.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.sensor.paired_sensor_status("ABCDE1234567")

        assert (
            str(err.value)
            == "SENSOR_PAIRED_SENSOR_STATUS command failed: sensor_error_loading"
        )


@pytest.mark.asyncio
async def test_paired_sensor_status_invalid_uid(
    mock_datagram_client: MagicMock,
) -> None:
    """Test that an invalid UID throws an exception.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.sensor.paired_sensor_status("$@&*!@--")

        assert str(err.value) == (
            "Invalid parameters provided: String is not alphanumeric for dictionary "
            "value @ data['uid']"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response",
    [load_fixture("paired_sensor_status_success_response.json").encode()],
)
async def test_paired_sensor_status_dump_success(
    mock_datagram_client: MagicMock,
) -> None:
    """Test the pair_dump command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            paired_sensor_status = await client.sensor.paired_sensor_status(
                "ABCDE1234567"
            )

        assert paired_sensor_status["command"] == 51
        assert paired_sensor_status["status"] == "ok"
        assert paired_sensor_status["data"] == {
            "uid": "6309FB799CDE",
            "codename": "gld1",
            "temperature": 68,
            "wet": False,
            "moved": True,
            "battery_percentage": 79,
        }



"""Define exception types for ``aioguardian``."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from aioguardian.helpers.command import Command

ERROR_CODE_MAPPING = {
    3: "sensor_not_paired",
    5: "sensor_error_loading",
    17: "valve_already_opened",
    18: "valve_already_closed",
    19: "valve_already_stopped",
    20: "valve_moving",
}


class GuardianError(Exception):
    """Define a base error from which all others inherit."""

    pass


class CommandError(GuardianError):
    """Define an error related to commands (invalid commands, invalid params, etc.)."""

    pass


class SocketError(GuardianError):
    """Define an error related to UDP socket issues."""

    pass


def _raise_on_command_error(command: Command, data: dict[str, Any]) -> None:
    """Examine a data response and raise errors appropriately.

    Args:
        command: The command to execute.
        data: An API response payload.o

    Raises:
        CommandError: Raised when a command fails for any reason.
    """
    if data.get("status") == "ok":
        return

    # If we know exactly why the command failed, raise that error:
    if data.get("error_code") in ERROR_CODE_MAPPING:
        raise CommandError(
            f"{command.name} command failed: {ERROR_CODE_MAPPING[data['error_code']]}"
        )

    # Last resort, return a generic error with the response payload:
    raise CommandError(f"{command.name} command failed (response: {data})")



"""Test the ununpair_sensor command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("unpair_sensor_failure_response.json").encode()]
)
async def test_unpair_sensor_failure(mock_datagram_client: MagicMock) -> None:
    """Test the unpair_sensor command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.sensor.unpair_sensor("abc123")

        assert str(err.value) == (
            "SENSOR_UNPAIR_SENSOR command failed "
            "(response: {'command': 50, 'status': 'error'})"
        )


@pytest.mark.asyncio
async def test_unpair_sensor_invalid_uid(mock_datagram_client: MagicMock) -> None:
    """Test that an invalid UID throws an exception.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.sensor.unpair_sensor("$@&*!@--")

        assert str(err.value) == (
            "Invalid parameters provided: String is not alphanumeric for dictionary "
            "value @ data['uid']"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("unpair_sensor_success_response.json").encode()]
)
async def test_unpair_sensor_success(mock_datagram_client: MagicMock) -> None:
    """Test the unpair_sensor command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            unpair_sensor_response = await client.sensor.unpair_sensor("abc123")

        assert unpair_sensor_response["command"] == 50
        assert unpair_sensor_response["status"] == "ok"



"""Run an example script to quickly test the guardian."""
import asyncio
import logging

from aioguardian import Client
from aioguardian.errors import GuardianError

_LOGGER = logging.getLogger(__name__)


async def main() -> None:
    """Create the aiohttp session and run the example."""
    logging.basicConfig(level=logging.INFO)

    async with Client("172.16.11.208") as guardian:
        try:
            publish_state_response = await guardian.iot.publish_state()
            _LOGGER.info("publish_state command response: %s", publish_state_response)
        except GuardianError as err:
            _LOGGER.info(err)


asyncio.run(main())



"""Test the halt command."""
from unittest.mock import MagicMock, Mock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("valve_halt_failure_response.json").encode()]
)
async def test_halt_failure(mock_datagram_client: MagicMock) -> None:
    """Test the valve_halt command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.valve.halt()

        assert str(err.value) == "VALVE_HALT command failed: valve_already_stopped"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("valve_halt_success_response.json").encode()]
)
async def test_halt_success(caplog: Mock, mock_datagram_client: MagicMock) -> None:
    """Test the valve_halt command succeeding.

    Args:
        caplog: A mocked logging facility.
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            valve_halt_response = await client.valve.halt()

        assert valve_halt_response["command"] == 19
        assert any(
            "The device will not respond to leak events" in e.message
            for e in caplog.records
        )



"""Test the onboard_sensor_status command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response",
    [load_fixture("onboard_sensor_status_success_response.json").encode()],
)
async def test_onboard_sensor_status_success(mock_datagram_client: MagicMock) -> None:
    """Test the onboard_sensor_status command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            onboard_sensor_status = await client.system.onboard_sensor_status()

        assert onboard_sensor_status["command"] == 80
        assert onboard_sensor_status["status"] == "ok"
        assert onboard_sensor_status["data"] == {"temperature": 71, "wet": False}


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response",
    [load_fixture("onboard_sensor_status_failure_response.json").encode()],
)
async def test_onboard_sensor_status_failure(mock_datagram_client: MagicMock) -> None:
    """Test the onboard_sensor_status command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.system.onboard_sensor_status()

        assert str(err.value) == (
            "SYSTEM_ONBOARD_SENSOR_STATUS command failed "
            "(response: {'command': 80, 'status': 'error'})"
        )



"""Test generic client characteristics."""
import asyncio
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from aioguardian import Client
from aioguardian.errors import SocketError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize("recv_response", [AsyncMock(side_effect=asyncio.TimeoutError)])
async def test_command_timeout(mock_datagram_client: MagicMock) -> None:
    """Test that a timeout during command execution throws an exception.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client, patch("asyncio.sleep"):
        with pytest.raises(SocketError) as err:
            async with Client("192.168.1.100") as client:
                await client.system.ping()

        assert str(err.value) == "SYSTEM_PING command timed out"


@pytest.mark.asyncio
async def test_command_timeout_successful_retry(
    mock_datagram_client: MagicMock,
) -> None:
    """Test that a timeout with a successful retry works.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client, patch("asyncio.sleep"):
        mock_datagram_client.recv.side_effect = [
            asyncio.TimeoutError,
            (load_fixture("ping_success_response.json").encode(), "192.168.1.100"),
        ]

        async with Client("192.168.1.100") as client:
            ping_response = await client.system.ping()

        assert ping_response["command"] == 0
        assert ping_response["status"] == "ok"
        assert ping_response["data"]["uid"] == "ABCDEF123456"


@pytest.mark.asyncio
async def test_command_without_socket_connect() -> None:
    """Test that executing a command without an open connection throws an exception."""
    client = Client("192.168.1.100")
    with pytest.raises(SocketError) as err:
        await client.system.ping()

    assert str(err.value) == "You aren't connected to the device yet"


@pytest.mark.asyncio
async def test_connect_timeout() -> None:
    """Test that a timeout during connection throws an exception."""
    with patch("asyncio_dgram.connect", AsyncMock(side_effect=asyncio.TimeoutError)):
        with pytest.raises(SocketError) as err:
            async with Client("192.168.1.100") as client:
                await client.system.ping()

        assert str(err.value) == "Connection to device timed out"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("ping_success_response.json").encode()]
)
async def test_raw_command_success(mock_datagram_client: MagicMock) -> None:
    """Test a successful raw command.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            ping_response = await client.execute_raw_command(0)

        assert ping_response["command"] == 0
        assert ping_response["status"] == "ok"
        assert ping_response["data"]["uid"] == "ABCDEF123456"



"""Run an example script to quickly test the guardian."""
import asyncio
import logging

from aioguardian import Client
from aioguardian.errors import GuardianError

_LOGGER = logging.getLogger(__name__)


async def main() -> None:
    """Create the aiohttp session and run the example."""
    logging.basicConfig(level=logging.INFO)

    async with Client("172.16.11.208") as guardian:
        try:
            diagnostics_response = await guardian.system.diagnostics()
            _LOGGER.info("diagnostics command response: %s", diagnostics_response)

            onboard_sensor_status = await guardian.system.onboard_sensor_status()
            _LOGGER.info(
                "onboard_sensor_status command response: %s", onboard_sensor_status
            )

            ping_response = await guardian.system.ping()
            _LOGGER.info("ping command response: %s", ping_response)
        except GuardianError as err:
            _LOGGER.info(err)


asyncio.run(main())



"""Run an example script to quickly test the guardian."""
import asyncio
import logging

from aioguardian import Client
from aioguardian.errors import GuardianError

_LOGGER = logging.getLogger(__name__)


async def main() -> None:
    """Create the aiohttp session and run the example."""
    logging.basicConfig(level=logging.INFO)

    async with Client("172.16.11.208") as guardian:
        try:
            valve_status_response = await guardian.valve.status()
            _LOGGER.info(
                "valve_status_response command response: %s", valve_status_response
            )

            valve_open_response = await guardian.valve.open()
            _LOGGER.info(
                "valve_open_response command response: %s", valve_open_response
            )

            # Give the valve a chance to open fully so that the valve_close command
            # doesn't error out:
            await asyncio.sleep(3)

            valve_close_response = await guardian.valve.close()
            _LOGGER.info(
                "valve_close_response command response: %s", valve_close_response
            )
        except GuardianError as err:
            _LOGGER.info(err)


asyncio.run(main())



"""Define WiFi commands."""
from collections.abc import Awaitable, Callable
from typing import Any

import voluptuous as vol

from aioguardian.errors import CommandError
from aioguardian.helpers.command import Command

PARAM_PASSWORD = "password"  # noqa: S105, # nosec
PARAM_SSID = "ssid"

WIFI_CONFIGURE_PARAM_SCHEMA = vol.Schema(
    {
        vol.Required(PARAM_SSID): vol.All(str, vol.Length(max=36)),
        vol.Required(PARAM_PASSWORD): vol.All(str, vol.Length(max=64)),
    }
)


class WiFiCommands:
    """Define an object to manage WiFi commands.

    Note that this class shouldn't be instantiated directly; an instance of it will
    automatically be added to the :meth:`Client <aioguardian.Client>` (as
    ``client.wifi``).

    Args:
        execute_command: The execute_command method from the Client object.
    """

    def __init__(
        self, execute_command: Callable[..., Awaitable[dict[str, Any]]]
    ) -> None:
        """Initialize.

        Args:
            execute_command: The execute_command method from the Client object.
        """
        self._execute_command = execute_command

    async def configure(
        self, ssid: str, password: str, *, silent: bool = True
    ) -> dict[str, Any]:
        """Configure the device to use a wireless network.

        Args:
            ssid: An SSID to connect to.
            password: The password to use to connect to the SSID.
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.

        Raises:
            CommandError: Raised when invalid parameters are provided.
        """
        params = {PARAM_SSID: ssid, PARAM_PASSWORD: password}

        try:
            WIFI_CONFIGURE_PARAM_SCHEMA(params)
        except vol.Invalid as err:
            raise CommandError(f"Invalid parameters provided: {err}") from None

        return await self._execute_command(
            Command.WIFI_CONFIGURE, params=params, silent=silent
        )

    async def disable_ap(self, *, silent: bool = True) -> dict[str, Any]:
        """Disable the device's onboard WiFi access point.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.WIFI_DISABLE_AP, silent=silent)

    async def enable_ap(self, *, silent: bool = True) -> dict[str, Any]:
        """Enable the device's onboard WiFi access point.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.WIFI_ENABLE_AP, silent=silent)

    async def list(self, *, silent: bool = True) -> dict[str, Any]:
        """List previously scanned nearby SSIDs.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.WIFI_LIST, silent=silent)

    async def reset(self, *, silent: bool = True) -> dict[str, Any]:
        """Erase and reset all WiFi settings.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.WIFI_RESET, silent=silent)

    async def scan(self, *, silent: bool = True) -> dict[str, Any]:
        """Scan for nearby SSIDs.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.WIFI_SCAN, silent=silent)

    async def status(self, *, silent: bool = True) -> dict[str, Any]:
        """Return the current WiFi status of the device.

        Args:
            silent: Whether the valve controller should beep upon successful command.

        Returns:
            An API response payload.
        """
        return await self._execute_command(Command.WIFI_STATUS, silent=silent)



"""Test the pair_dump command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("pair_dump_failure_response.json").encode()]
)
async def test_pair_dump_failure(mock_datagram_client: MagicMock) -> None:
    """Test the pair_dump command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.sensor.pair_dump()

        assert str(err.value) == (
            "SENSOR_PAIR_DUMP command failed "
            "(response: {'command': 48, 'status': 'error'})"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("pair_dump_success_response.json").encode()]
)
async def test_pair_dump_success(mock_datagram_client: MagicMock) -> None:
    """Test the pair_dump command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            pair_dump_response = await client.sensor.pair_dump()

        assert pair_dump_response["command"] == 48
        assert pair_dump_response["status"] == "ok"
        assert pair_dump_response["data"] == {
            "pair_count": 1,
            "paired_uids": ["6309FB799CDE"],
        }



"""Run an example script to quickly test the guardian."""
import asyncio
import logging

from aioguardian import Client
from aioguardian.errors import GuardianError

_LOGGER = logging.getLogger(__name__)


async def main() -> None:
    """Create the aiohttp session and run the example."""
    logging.basicConfig(level=logging.INFO)

    async with Client("172.16.11.208") as guardian:
        try:
            pair_dump_response = await guardian.sensor.pair_dump()
            _LOGGER.info("pair_dump_response command response: %s", pair_dump_response)

            for uid in pair_dump_response["data"]["paired_uids"]:
                paired_sensor_status_resp = await guardian.sensor.paired_sensor_status(
                    uid
                )
                _LOGGER.info(
                    "paired_sensor_status command response (UID: %s): %s",
                    uid,
                    paired_sensor_status_resp,
                )
        except GuardianError as err:
            _LOGGER.info(err)


asyncio.run(main())



"""Define custom validators."""
from typing import Any, cast
from urllib.parse import urlparse

import voluptuous as vol


def alphanumeric(value: Any) -> str:
    """Validate that a string is alphanumeric only.

    Args:
        value: A value to evaluae.

    Returns:
        An alphanumeric string.

    Raises:
        Invalid: Raised when a non-alphanumeric string is encountered.
    """
    str_value = str(value)

    if str_value.isalnum():
        return str_value

    raise vol.Invalid("String is not alphanumeric")


def url(value: Any) -> str:
    """Validate that a string is a URL.

    Args:
        value: A value to evaluae.

    Returns:
        An URL string.

    Raises:
        Invalid: Raised when a non-URL string is encountered.
    """
    url_in = str(value)

    if urlparse(url_in).scheme in ["http", "https"]:
        return cast(
            str,
            vol.Schema(vol.Url())(url_in),  # pylint: disable=no-value-for-parameter
        )

    raise vol.Invalid("Invalid URL")



"""Test the diagnostics command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("diagnostics_failure_response.json").encode()]
)
async def test_diagnostics_failure(mock_datagram_client: MagicMock) -> None:
    """Test the diagnostics command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.system.diagnostics()

        assert str(err.value) == (
            "SYSTEM_DIAGNOSTICS command failed "
            "(response: {'command': 1, 'status': 'error'})"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("diagnostics_success_response.json").encode()]
)
async def test_diagnostics_success(mock_datagram_client: MagicMock) -> None:
    """Test the diagnostics command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            diagnostics_response = await client.system.diagnostics()

        assert diagnostics_response["command"] == 1
        assert diagnostics_response["status"] == "ok"
        assert diagnostics_response["data"] == {
            "codename": "gvc1",
            "uid": "ABCDEF123456",
            "uptime": 41,
            "firmware": "0.20.9-beta+official.ef3",
            "rf_modem_firmware": "4.0.0",
            "available_heap": 34456,
        }



"""Define a client object to interact with a Guardian device."""
from __future__ import annotations

import asyncio
import json
from types import TracebackType
from typing import Any, cast

import asyncio_dgram

from aioguardian.commands.iot import IOTCommands
from aioguardian.commands.sensor import SensorCommands
from aioguardian.commands.system import SystemCommands
from aioguardian.commands.valve import ValveCommands
from aioguardian.commands.wifi import WiFiCommands
from aioguardian.const import LOGGER
from aioguardian.errors import SocketError, _raise_on_command_error
from aioguardian.helpers.command import Command, get_command_from_code

DEFAULT_COMMAND_RETRIES: int = 3
DEFAULT_PORT: int = 7777
DEFAULT_REQUEST_TIMEOUT: int = 10

try:
    timeout_callable = asyncio.timeout  # type: ignore[attr-defined]
except AttributeError:  # pragma: no cover
    import async_timeout  # pylint: disable=import-error

    timeout_callable = async_timeout.timeout


class Client:  # pylint: disable=too-many-instance-attributes
    """Define the class that can send commands to a Guardian device.

    Args:
        ip_address: The IP address or hostname of a Guardian valve controller.
        port: The port to connect to.
        request_timeout: The number of seconds to wait before timing out a request.
        command_retries: The number of retries to use on a failed command.
    """

    def __init__(
        self,
        ip_address: str,
        *,
        port: int = DEFAULT_PORT,
        request_timeout: int = DEFAULT_REQUEST_TIMEOUT,
        command_retries: int = DEFAULT_COMMAND_RETRIES,
    ) -> None:
        """Initialize.

        Args:
            ip_address: The IP address or hostname of a Guardian valve controller.
            port: The port to connect to.
            request_timeout: The number of seconds to wait before timing out a request.
            command_retries: The number of retries to use on a failed command.
        """
        self._command_retries = command_retries
        self._ip = ip_address
        # Since device communication happens over a single UDP port, concurrent
        # operations can return faulty data; we use a lock so the user doesn't have to
        # know anything about that:
        self._lock: asyncio.Lock = asyncio.Lock()
        self._port = port
        self._request_timeout = request_timeout
        self._stream: asyncio_dgram.aio.DatagramStream | None = None

        self.iot = IOTCommands(self._execute_command)
        self.sensor = SensorCommands(self._execute_command)
        self.system = SystemCommands(self._execute_command)
        self.valve = ValveCommands(self._execute_command)
        self.wifi = WiFiCommands(self._execute_command)

    async def __aenter__(self) -> Client:
        """Define an entry point into this object via a context manager.

        Returns:
            A connected aioguardian client.
        """
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,  # noqa: F841
        exc_val: BaseException | None,  # noqa: F841
        exc_tb: TracebackType | None,  # noqa: F841
    ) -> None:
        """Define an exit point out of this object via a context manager.

        Args:
            exc_type: An optional exception if one caused the context manager to close.
            exc_val: The value of the optional exception
            exc_tb: The traceback of the optional exception
        """
        self.disconnect()

    async def _execute_command(
        self, command: Command, *, params: dict | None = None, silent: bool = True
    ) -> dict[str, Any]:
        """Make a request against the Guardian device and return the response.

        Args:
            command: The command to execute.
            params: Any parameters to send along with the command.
            silent: If ``True``, silence "beep" tones associated with this command.

        Returns:
            An API response payload.

        Raises:
            SocketError: Raised on an issue with the UDP socket.
        """
        if not self._stream:
            raise SocketError("You aren't connected to the device yet")

        _params = params or {}
        payload = {"command": command.value, "silent": silent, **_params}

        retry = 0

        while retry < self._command_retries:
            try:
                async with self._lock, timeout_callable(self._request_timeout):
                    await self._stream.send(  # type: ignore[attr-defined]
                        json.dumps(payload).encode()
                    )
                    data, remote_addr = await self._stream.recv()
                    break
            except asyncio.TimeoutError:
                LOGGER.info("%s command timed out; trying again", command.name)
                retry += 1
                await asyncio.sleep(1)
        else:
            raise SocketError(f"{command.name} command timed out")

        decoded_data = json.loads(data.decode())
        LOGGER.debug("Received data from %s: %s", remote_addr, decoded_data)

        _raise_on_command_error(command, decoded_data)

        return cast(dict[str, Any], decoded_data)

    async def connect(self) -> None:
        """Connect to the Guardian device.

        Raises:
            SocketError: Raised on an issue with the UDP socket.
        """
        try:
            async with timeout_callable(self._request_timeout):
                self._stream = await asyncio_dgram.connect((self._ip, self._port))
        except asyncio.TimeoutError:
            raise SocketError("Connection to device timed out") from None

    def disconnect(self) -> None:
        """Close the connection."""
        if self._stream:
            self._stream.close()
            self._stream = None

    async def execute_raw_command(
        self, command_code: int, *, params: dict | None = None, silent: bool = True
    ) -> dict[str, Any]:
        """Execute a command via its integer-based command code.

        Args:
            command_code: The command code to execute.
            params: Any parameters to send along with the command.
            silent: If ``True``, silence "beep" tones associated with this command.

        Returns:
            An API response payload.
        """
        command = get_command_from_code(command_code)
        return await self._execute_command(command, params=params, silent=silent)



"""Test the list command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("wifi_list_failure_response.json").encode()]
)
async def test_list_failure(mock_datagram_client: MagicMock) -> None:
    """Test the wifi_list command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.wifi.list()

        assert str(err.value) == (
            "WIFI_LIST command failed (response: {'command': 38, 'status': 'error'})"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("wifi_list_success_response.json").encode()]
)
async def test_list_success(mock_datagram_client: MagicMock) -> None:
    """Test the wifi_list command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            await client.wifi.scan()
            wifi_list_response = await client.wifi.list()

        assert wifi_list_response["command"] == 38
        assert wifi_list_response["status"] == "ok"
        assert wifi_list_response["data"] == {
            "record_count": 1,
            "records": [
                {
                    "bssid": "60:31:97:BE:53:5D",
                    "rssi": -89,
                    "channel": 1,
                    "authmode": 4,
                    "ssid": "CenturyLink0526",
                }
            ],
        }


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("wifi_list_success_response.json").encode()]
)
async def test_list_success_cached(mock_datagram_client: MagicMock) -> None:
    """Test the wifi_list command succeeding after a cache.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            wifi_list_response = await client.wifi.list()

        assert wifi_list_response["command"] == 38
        assert wifi_list_response["status"] == "ok"
        assert wifi_list_response["data"] == {
            "record_count": 1,
            "records": [
                {
                    "bssid": "60:31:97:BE:53:5D",
                    "rssi": -89,
                    "channel": 1,
                    "authmode": 4,
                    "ssid": "CenturyLink0526",
                }
            ],
        }



"""Test the reset command."""
from unittest.mock import MagicMock

import pytest

from aioguardian import Client
from aioguardian.errors import CommandError
from tests.common import load_fixture


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("valve_reset_failure_response.json").encode()]
)
async def test_reset_failure(mock_datagram_client: MagicMock) -> None:
    """Test the valve_reset command failing.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        with pytest.raises(CommandError) as err:
            async with Client("192.168.1.100") as client:
                _ = await client.valve.reset()

        assert (
            str(err.value) == "VALVE_RESET command failed "
            "(response: {'command': 20, 'status': 'error'})"
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "command_response", [load_fixture("valve_reset_success_response.json").encode()]
)
async def test_reset_success(mock_datagram_client: MagicMock) -> None:
    """Test the valve_reset command succeeding.

    Args:
        mock_datagram_client: A mocked UDP client.
    """
    with mock_datagram_client:
        async with Client("192.168.1.100") as client:
            valve_reset_response = await client.valve.reset()

        assert valve_reset_response["command"] == 20
